/*
 *  linux/fs/qnx4fs/bitmap.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  14 jan 2004 Dmitry Yakovlev - 2.6 kernel support
 *  14 mar 2004 Dmitry Yakovlev - bitmap protection
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/fs.h>
#include "qnx4fs_fs.h"
#include <linux/buffer_head.h>

static inline void
clear_bits(u8 *ptr, int from, int len)
{
	unsigned char dat;
	dat = ~(0x0FF << len) << from;
	if ((*ptr & dat) != dat)
		QNX4FS_WARN(" bits are not set\n");
	*ptr &= ~dat;
}

static inline void
clear_low_bits(u8 *ptr, int len)
{
	if ((*ptr & (~(0xff << len))) != ((~(0xff << len)) & 0xff))
		QNX4FS_WARN(" bits are not set\n");
	*ptr &= (0xff << len);
}

static inline void
clear_high_bits(u8 *ptr, int len)
{
	if ((*ptr & ~(0xff >> len)) != (~(0xff >> len) & 0xff))
		QNX4FS_WARN(" bits are not set\n");
	*ptr &= (0xff >> len);
}

static inline void
clear_bytes(u8 *bitmap, int from, int to)
{
	for (bitmap += from, to -= from; to > 0; to--) {
		if (*bitmap != 0xff)
			QNX4FS_WARN(" byte is not set\n");
		*bitmap++ = 0;
	}
}

static inline struct buffer_head *
bitmap_sb_bread(struct super_block *sb, unsigned long block)
{
	struct buffer_head *bh;
	if (block > QNX4FS_SB(sb)->s_bitmap_blocks) {
		QNX4FS_WARN(" wrong bitmap block %lu\n", block);
		return NULL;
	}
	bh = sb_bread(sb, block + QNX4FS_BITMAP_BLOCK - 1);
	if (!bh)
		QNX4FS_WARN(" unable to read bitmap block %lu\n", block);
	return bh;
}

static inline void
clear_sb_blocks(struct super_block *sbp, unsigned long from, unsigned long to)
{
	struct buffer_head *bp;
	unsigned char *bpt;
	int i;

	for (; from < to; from++) {
		bp = bitmap_sb_bread(sbp, from);
		if (bp == NULL)
			continue;
		bpt = (unsigned char *)(bp->b_data);
		mutex_lock(&QNX4FS_SB(sbp)->s_bmlock);
		for (i = 0; i < QNX4FS_BLOCK_SIZE; i++) {
			if (*bpt++ != 0xFF) {
				QNX4FS_WARN(" byte is not set\n");
				break;
			}
		}
		memset(bp->b_data, 0, QNX4FS_BLOCK_SIZE);
		mark_buffer_dirty(bp);
		mutex_unlock(&QNX4FS_SB(sbp)->s_bmlock);
		brelse(bp);
	}
}

void
qnx4fs_bfree(struct super_block *sbp, unsigned long from, unsigned long len)
{
	struct buffer_head *bp;
	unsigned long to;
	unsigned int bfrom, bto;
	unsigned long block_from, block_to;
	u8 *bitmap;

	if (len == 0 || len > QNX4FS_SB(sbp)->s_blocks) {
		if (len)
			QNX4FS_WARN(" lerror from %ld len %lu\n", from, len);
		return;
	}

	to = from + len - 1;
	/*
	 * we will never free (boot, root, bitmap, rootdir) blocks)
	 */
	if ((to > QNX4FS_SB(sbp)->s_blocks) ||
		(from < (QNX4FS_SB(sbp)->s_bitmap_blocks + 6))) {
		QNX4FS_WARN(" error from %ld len %lu\n", from, len);
		return;
	}

	from--;

	block_from = from >> (QNX4FS_BLOCK_SIZE_BITS + 3);
	block_to   = to >> (QNX4FS_BLOCK_SIZE_BITS + 3);
	from	  &= (QNX4FS_BLOCK_SIZE * 8 - 1);
	to	  &= (QNX4FS_BLOCK_SIZE * 8 - 1);
	bfrom      = from >> 3;
	bto        = to >> 3;

	bp = bitmap_sb_bread(sbp, block_to);
	if (bp == NULL)
		return;
	bitmap = (u8 *)bp->b_data;
	switch (block_to - block_from) {
	case 0:
		switch (bto - bfrom) {
		case 0:
			mutex_lock(&QNX4FS_SB(sbp)->s_bmlock);
			clear_bits(&bitmap[bfrom], from & 7, len);
			mutex_unlock(&QNX4FS_SB(sbp)->s_bmlock);
			break;
		default:
			mutex_lock(&QNX4FS_SB(sbp)->s_bmlock);
			clear_low_bits(&bitmap[bto], to & 7);
			clear_bytes(bitmap, (from + 7) >> 3, bto);
			if (from & 7)
				clear_high_bits(&bitmap[bfrom], 8 - (from & 7));
			mutex_unlock(&QNX4FS_SB(sbp)->s_bmlock);
			break;
		}
		break;
	default:
		mutex_lock(&QNX4FS_SB(sbp)->s_bmlock);
		clear_low_bits(&bitmap[bto], to & 7);
		clear_bytes(bitmap, 0, bto);
		mutex_unlock(&QNX4FS_SB(sbp)->s_bmlock);
		mark_buffer_dirty(bp);
		brelse(bp);
		clear_sb_blocks(sbp, block_from + 1, block_to);
		bp = bitmap_sb_bread(sbp, block_from);
		if (bp == NULL)
			return;
		mutex_lock(&QNX4FS_SB(sbp)->s_bmlock);
		bitmap = (u8 *)bp->b_data;
		clear_bytes(bitmap, (from + 7) >> 3, QNX4FS_BLOCK_SIZE);
		if (from & 7)
			clear_high_bits(&bitmap[bfrom], 8 - (from & 7));
		mutex_unlock(&QNX4FS_SB(sbp)->s_bmlock);
		break;
	}
	mark_buffer_dirty(bp);
	brelse(bp);
}

int
qnx4fs_balloc(struct super_block *sbp, unsigned long *goal)
{
	struct buffer_head *bp;
	unsigned long block, nblocks;
	int nr, res;
	void *dp;

	block = (*goal - 1) >> (QNX4FS_BLOCK_SIZE_BITS + 3);
	nr = (*goal - 1) & (QNX4FS_BLOCK_SIZE * 8 - 1);
	bp = bitmap_sb_bread(sbp, block);
	if (bp) {
		mutex_lock(&QNX4FS_SB(sbp)->s_bmlock);
		dp = bp->b_data;
		if (test_bit(nr, dp)) {
			mutex_unlock(&QNX4FS_SB(sbp)->s_bmlock);
			brelse(bp);
		} else {
			res = 1;
			goto found;
		}
	}
again:
	block = 0;
	nr = 0;
	res = 0;
	nblocks = QNX4FS_SB(sbp)->s_bitmap_blocks;
	for (; block < nblocks; block++, nr = 0) {
		bp = bitmap_sb_bread(sbp, block);
		if (bp == NULL)
			continue;
		dp = bp->b_data;
		mutex_lock(&QNX4FS_SB(sbp)->s_bmlock);
		nr = find_next_zero_bit(dp, QNX4FS_BLOCK_SIZE * 8, nr);
		if (nr < QNX4FS_BLOCK_SIZE * 8) {
			*goal = (block << (QNX4FS_BLOCK_SIZE_BITS+3)) + 1 + nr;
			goto found;
		}
		mutex_unlock(&QNX4FS_SB(sbp)->s_bmlock);
		brelse(bp);
	}
	return -ENOSPC;
found:
	if (test_and_set_bit(nr, (unsigned long *)dp)) {
		QNX4FS_WARN(" bit set nr %d, block %lu, res %d \n", nr, block,
			res);
		mutex_unlock(&QNX4FS_SB(sbp)->s_bmlock);
		brelse(bp);
		goto again;
	}
	mutex_unlock(&QNX4FS_SB(sbp)->s_bmlock);
	mark_buffer_dirty(bp);
	brelse(bp);
	return res;
}

unsigned long
qnx4fs_count_free_blocks(struct super_block *sbp)
{
	struct buffer_head *bp;
	unsigned long count = 0;
	unsigned long nblocks, block;
	unsigned *w;
	int i;

	nblocks = QNX4FS_SB(sbp)->s_bitmap_blocks;
	for (block = 0; block < nblocks; block++) {
		bp = bitmap_sb_bread(sbp, block);
		if (bp == NULL)
			continue;
		w = (unsigned *)bp->b_data;
		for (i = QNX4FS_BLOCK_SIZE/4; i > 0; i--)
			count += hweight32(*w++);
		brelse(bp);
	}
	return (QNX4FS_SB(sbp)->s_bitmap_blocks
		<< (QNX4FS_BLOCK_SIZE_BITS + 3)) - count;
}
