/*
 *  linux/fs/qnx4fs/file.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  14 jan 2004 Dmitry Yakovlev - 2.6 kernel support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/buffer_head.h>
#include "qnx4fs_fs.h"

unsigned long
qnx4fs_map_block(struct inode *ip, long block)
{
	struct buffer_head *bp;
	struct qnx4fs_xb *xb;
	struct qnx4fs_xt *xt;
	unsigned long phys;
	long xbnr, offset;
	int xtnr, nxt;

	xbnr = QNX4FS_I(ip)->i_xbnr;
	xtnr = QNX4FS_I(ip)->i_xtnr;
	if (block >= QNX4FS_I(ip)->i_offset) {
		if (block < QNX4FS_I(ip)->i_offset_end) {
			phys = (QNX4FS_I(ip)->i_phys + block -
				 QNX4FS_I(ip)->i_offset);
			goto found1;
		}
		for (offset = QNX4FS_I(ip)->i_offset_end;
			    xbnr;
			    xbnr = le32_to_cpu(xb->xb_next),
			    brelse(bp),
			    xtnr = -1) {

			bp = qnx4fs_sb_bread(ip->i_sb, xbnr - 1);
			if (bp == NULL) {
				QNX4FS_WARN(" Wrong xblk xbnr %ld\n", xbnr);
				goto outerr;
			}
			xb = (struct qnx4fs_xb *)bp->b_data;
			if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
				brelse(bp);
				QNX4FS_WARN(" Wrong xblk signature xbnr %ld\n",
					xbnr);
				goto outerr;
			}
			nxt = xb->xb_nxt;
			if (xtnr >= nxt - 1)
				continue;
			xt = &xb->xb_xts[++xtnr];
			for (; xtnr < nxt; xtnr++, xt++) {
				offset += le32_to_cpu(xt->size);
				if (block < offset) {
					phys = le32_to_cpu(xt->start)
					       + le32_to_cpu(xt->size)
					       + block - offset;
					QNX4FS_I(ip)->i_offset = offset -
						le32_to_cpu(xt->size);
					QNX4FS_I(ip)->i_offset_end = offset;
					goto found;
				}
			}
		};
	} else if (block < QNX4FS_I(ip)->i_xt0.size) {
		phys = (QNX4FS_I(ip)->i_xt0.start + block);
		goto found1;
	}
	for (offset = QNX4FS_I(ip)->i_offset;
		    xbnr;
		    xbnr = le32_to_cpu(xb->xb_prev),
		    brelse(bp),
		    xtnr = QNX4FS_MAX_XTS_PER_XBLK) {

		bp = qnx4fs_sb_bread(ip->i_sb, xbnr - 1);
		if (bp == NULL) {
			QNX4FS_WARN(" wrong xblk xbnr %ld\n", xbnr);
			goto outerr;
		}
		xb = (struct qnx4fs_xb *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			brelse(bp);
			QNX4FS_WARN(" wrong xblk signature xbnr %ld\n", xbnr);
			goto outerr;
		}
		if (xtnr <= 0)
			continue;
		nxt = xb->xb_nxt;
		xt = &xb->xb_xts[--xtnr];
		for (; xtnr >= 0; xtnr--, xt--) {
			offset -= le32_to_cpu(xt->size);
			if (block >= offset) {
				phys = le32_to_cpu(xt->start) + block - offset;
				QNX4FS_I(ip)->i_offset = offset;
				QNX4FS_I(ip)->i_offset_end = offset +
					le32_to_cpu(xt->size);
				goto found;
			}
		}
	};
outerr:
	QNX4FS_WARN(" inode %ld, block %d not found\n", ip->i_ino, (int)block);
	return 0;
found:
	QNX4FS_I(ip)->i_phys = le32_to_cpu(xt->start);
	QNX4FS_I(ip)->i_xtnr = xtnr;
	QNX4FS_I(ip)->i_xbnr = xbnr;
	brelse(bp);
found1:
	QNX4FS_DEBUG(" inode %ld, block %d, phys %d\n", ip->i_ino, (int)block,
		(int)phys);
	return phys;
}

int
qnx4fs_get_block(struct inode *ip, sector_t block,
		  struct buffer_head *bh_result, int create)
{
	unsigned long phys;
	long nblocks;

	nblocks = (QNX4FS_I(ip)->i_mmu_private + QNX4FS_BLOCK_SIZE - 1)
		>> QNX4FS_BLOCK_SIZE_BITS;
	QNX4FS_DEBUG(" inode %ld, block %d, nblocks %ld, create %d\n",
		ip->i_ino, (int)block, nblocks, create);
	if (create && block >= nblocks) {

		phys = qnx4fs_iexpand(ip);
		if (!phys) {
			QNX4FS_WARN(" unable to expand ino %ld, block %d,"\
				"nblocks %ld\n", ip->i_ino, (int)block,
				nblocks);
			return -ENOSPC;
		}
		set_buffer_new(bh_result);

	} else {
		phys = qnx4fs_map_block(ip, block);
		if (!phys) {
			QNX4FS_WARN(" unable to map ino %ld, block %d, "\
			"nblocks %ld\n", ip->i_ino, (int)block, nblocks);

			return -EIO;
		}
	}
	map_bh(bh_result, ip->i_sb, phys - 1);
	return 0;
}

const struct file_operations qnx4fs_file_ops = {
	.llseek		= generic_file_llseek,
	.read		= do_sync_read,
	.aio_read	= generic_file_aio_read,
	.write		= do_sync_write,
	.aio_write	= generic_file_aio_write,
	.fsync		= file_fsync,
	.mmap		= generic_file_mmap,
};

struct inode_operations qnx4fs_file_inode_ops = {
	.truncate	= qnx4fs_truncate,
};

static int
qnx4fs_readpage(struct file *filp, struct page *pp)
{
	return block_read_full_page(pp, qnx4fs_get_block);
}

static int
qnx4fs_writepage(struct page *pp, struct writeback_control *wbc)
{
	QNX4FS_DEBUG(" inode %ld\n", pp->mapping->host->i_ino);
	return block_write_full_page(pp, qnx4fs_get_block, wbc);
}

static int
qnx4fs_write_begin(struct file *file, struct address_space *mapping,
			loff_t pos, unsigned len, unsigned flags,
			struct page **pagep, void **fsdata)
{
	*pagep = NULL;
	return cont_write_begin(file, mapping, pos, len, flags, pagep, fsdata,
				qnx4fs_get_block,
				&QNX4FS_I(mapping->host)->i_mmu_private);
}

static sector_t
qnx4fs_bmap(struct address_space *mapping, sector_t block)
{
	return generic_block_bmap(mapping, block, qnx4fs_get_block);
}

struct address_space_operations qnx4fs_address_space_ops = {
	.readpage	= qnx4fs_readpage,
	.writepage	= qnx4fs_writepage,
	.sync_page	= block_sync_page,
	.write_begin    = qnx4fs_write_begin,
	.write_end      = generic_write_end,
	.bmap		= qnx4fs_bmap,
};
