/*
 *  linux/fs/qnx4fs/ialloc.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  14 jan 2004 Dmitry Yakovlev - 2.6 kernel support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/pagemap.h>
#include <linux/fs.h>
#include "qnx4fs_fs.h"
#include <linux/buffer_head.h>


int
qnx4fs_dialloc(struct super_block *sbp, ino_t *inop)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;

	/* /.inodes file extension protection */
	mutex_lock(&QNX4FS_SB(sbp)->s_iflock);
	bp = qnx4fs_get_slot(QNX4FS_SB(sbp)->s_inodes, inop, &qdp,
			     QNX4FS_INODES_EXPAND_DELTA);

	if (IS_ERR(bp)) {
		mutex_unlock(&QNX4FS_SB(sbp)->s_iflock);
		return PTR_ERR(bp);
	}

	qnx4fs_mark_slot(qdp, bp, (QNX4FS_FILE_USED | QNX4FS_FILE_INODE));
	mutex_unlock(&QNX4FS_SB(sbp)->s_iflock);
	return 0;
}


void
qnx4fs_difree(struct inode *ip, ino_t ino)
{
	struct super_block *sbp = ip->i_sb;
	struct buffer_head *bp;
	qnx4fs_dinode_t *qdp;

	bp = qnx4fs_sb_bread(ip->i_sb, ino2phys(ino));
	if (bp == NULL) {
		QNX4FS_WARN(" inode %ld\n", ino);
		return;
	}
	mutex_lock(&QNX4FS_SB(sbp)->s_iflock);
	qdp = (qnx4fs_dirent_t *)bp->b_data + ino2slot(ino);
	qnx4fs_mark_slot(qdp, bp, 0);
	mutex_unlock(&QNX4FS_SB(sbp)->s_iflock);
}

struct buffer_head *
qnx4fs_inew(struct inode *dip, struct dentry *dp,
	     struct inode **ipp, qnx4fs_dirent_t **qdpp)
{
	struct buffer_head *bp;
	struct inode *ip;
	ino_t ino;
	int err, longname = (dp->d_name.len > QNX4FS_SHORT_NAME_MAX);

	QNX4FS_DEBUG(" name <%s>\n", dp->d_name.name);

	ip = new_inode(dip->i_sb);
	if (ip == NULL)
		return ERR_PTR(-ENOMEM);

	bp = qnx4fs_get_slot(dip, &ino, qdpp, QNX4FS_DIR_EXPAND_DELTA);
	if (IS_ERR(bp)) {
		qnx4fs_dec_nlink(ip);
		iput(ip);
		return bp;
	}

	if (longname) {
		err = qnx4fs_dialloc(dip->i_sb, &ino);
		if (err) {
			brelse(bp);
			qnx4fs_dec_nlink(ip);
			iput(ip);
			return ERR_PTR(err);
		}
	}

	QNX4FS_I(ip)->i_xt0.size 	= 0;
	QNX4FS_I(ip)->i_xt0.start 	= 0;
	QNX4FS_I(ip)->i_xbnr0 		= 0;
	QNX4FS_I(ip)->i_xbnr_last 	= 0;
	QNX4FS_I(ip)->i_offset		= 0;
	QNX4FS_I(ip)->i_offset_end	= 0;
	QNX4FS_I(ip)->i_phys		= 0;
	QNX4FS_I(ip)->i_xtnr		= -1;
	QNX4FS_I(ip)->i_xbnr		= 0;
	QNX4FS_I(ip)->i_linked		= longname;
	QNX4FS_I(ip)->i_nxt 		= 0;
	QNX4FS_I(ip)->i_mmu_private 	= 0;

	*ipp			= ip;
	ip->i_ino		= ino;
	ip->i_uid		= current_fsuid();
	ip->i_gid		= (dip->i_mode & S_ISGID)
				  ? dip->i_gid : current_fsgid();
	ip->i_mtime		= ip->i_atime
				= ip->i_ctime
				= CURRENT_TIME;
	ip->i_blocks		= 0;

	return bp;
}

static struct qnx4fs_xb *
qnx4fs_xalloc(struct inode *ip, struct buffer_head **bpp,
	       unsigned long *xbnrp)
{
	struct buffer_head *bp;
	struct qnx4fs_xb *xb;
	unsigned long goal = 1;

	QNX4FS_DEBUG(" inode %ld\n", ip->i_ino);

	if (qnx4fs_balloc(ip->i_sb, &goal)) {
		/* QNX4FS_WARN(" balloc %lu\n", goal); */
		return NULL;
	}

	bp = qnx4fs_sb_bread(ip->i_sb, goal - 1);

	if (bp == NULL) {
		qnx4fs_bfree(ip->i_sb, goal, 1);
		QNX4FS_WARN(" bread %lu\n", goal);
		return NULL;
	}
	xb = (struct qnx4fs_xb *)bp->b_data;
	memset(xb, 0, QNX4FS_BLOCK_SIZE);
	xb->xb_prev	= 0;
	xb->xb_next	= 0;
	xb->xb_nxt	= 1;
	memcpy(xb->xb_signature, "IamXblk\0", 8);
	xb->xb_xt0.start = 0;
	xb->xb_xt0.size  = 0;
	++ip->i_blocks;

	*bpp = bp;
	*xbnrp = goal;

	return xb;
}

struct buffer_head *
qnx4fs_iexpand_special(struct inode *ip, int n)
{
	struct buffer_head *bp, *rbp;
	unsigned long phys;
	int i;

	QNX4FS_DEBUG(" inode %ld, n %d\n", ip->i_ino, n);
	phys = qnx4fs_iexpand(ip);
	if (!phys)
		return NULL;
	rbp = qnx4fs_sb_bread(ip->i_sb, phys - 1);
	if (rbp == NULL) {
		QNX4FS_WARN(" unable1 to read inode %ld, phys %lu\n",
			ip->i_ino, phys);
		goto fail;
	}
	memset(rbp->b_data, 0, QNX4FS_BLOCK_SIZE);
	mark_buffer_dirty(rbp);
	ip->i_size += QNX4FS_BLOCK_SIZE;
	for (i = n - 1; i > 0; i--) {
		phys = qnx4fs_iexpand(ip);
		if (!phys)
			break;
		bp = qnx4fs_sb_bread(ip->i_sb, phys - 1);
		if (bp == NULL) {
			QNX4FS_WARN(" unable to read inode %ld, phys %lu\n",
				 ip->i_ino, phys);
			break;
		}
		memset(bp->b_data, 0, QNX4FS_BLOCK_SIZE);
		ip->i_size += QNX4FS_BLOCK_SIZE;
		mark_buffer_dirty(bp);
		brelse(bp);
	}
	mark_inode_dirty(ip);
	return rbp;
fail:
	/* QNX4FS_WARN(" inode %ld, n %d\n", ip->i_ino, n); */
	/* qnx4fs_truncate(ip); */
	return NULL;
}

unsigned long
qnx4fs_iexpand(struct inode *ip)
{
	struct buffer_head *bp = NULL, *bp2 = NULL;
	struct qnx4fs_xb *xb = NULL;
	unsigned long goal, tmp, xbnr = QNX4FS_I(ip)->i_xbnr_last;
	int ixt = 0, res;

	QNX4FS_DEBUG(" inode %ld\n", ip->i_ino);
	switch (QNX4FS_I(ip)->i_nxt) {
	case 0:
		goal = 1;
		break;
	case 1:
		goal = QNX4FS_I(ip)->i_xt0.start + QNX4FS_I(ip)->i_xt0.size;
		break;
	default:
		bp = qnx4fs_sb_bread(ip->i_sb, xbnr - 1);
		if (bp == NULL)
			goto fail;
		xb = (struct qnx4fs_xb *)bp->b_data;
		ixt = xb->xb_nxt - 1;
		/* extending last extent in last xblk */
		goal = le32_to_cpu(xb->xb_xts[ixt].start)
		     + le32_to_cpu(xb->xb_xts[ixt].size);
		break;
	}

	res = qnx4fs_balloc(ip->i_sb, &goal);

	if (res < 0) {
		if (bp != NULL)
			brelse(bp);
		QNX4FS_WARN(" balloc goal %lu\n", goal);
		goto fail;
	}

	switch (QNX4FS_I(ip)->i_nxt) {
	case 0:
		++QNX4FS_I(ip)->i_nxt;
		QNX4FS_I(ip)->i_xt0.start  = goal;
		QNX4FS_I(ip)->i_xt0.size   = 1;
		QNX4FS_I(ip)->i_phys       = goal;
		QNX4FS_I(ip)->i_offset_end = 1;
		break;
	case 1:
		if (res) {
			++QNX4FS_I(ip)->i_xt0.size;
			++QNX4FS_I(ip)->i_offset_end;
		} else {
			xb = qnx4fs_xalloc(ip, &bp, &xbnr);
			if (xb == NULL)
				goto fail_bfree;

			QNX4FS_I(ip)->i_xbnr0	= xbnr;
			QNX4FS_I(ip)->i_xbnr_last = xbnr;
			QNX4FS_I(ip)->i_xbnr	= xbnr;
			xb->xb_xts[0].start	= cpu_to_le32(goal);
			xb->xb_xts[0].size	= cpu_to_le32(1);
			xb->xb_nblocks		= cpu_to_le32(1);

			xb->xb_xt0.start
				= cpu_to_le32(QNX4FS_I(ip)->i_xt0.start);
			xb->xb_xt0.size
				= cpu_to_le32(QNX4FS_I(ip)->i_xt0.size);

			++QNX4FS_I(ip)->i_nxt;
		}
		break;
	default:
		tmp = le32_to_cpu(xb->xb_nblocks) + 1;
		if (res) {
			xb->xb_nblocks = cpu_to_le32(tmp);
			tmp = le32_to_cpu(xb->xb_xts[ixt].size) + 1;
			xb->xb_xts[ixt].size = cpu_to_le32(tmp);
			if (QNX4FS_I(ip)->i_phys ==
				le32_to_cpu(xb->xb_xts[ixt].start))
				++QNX4FS_I(ip)->i_offset_end;
			break;
		}

		if (++ixt >= QNX4FS_MAX_XTS_PER_XBLK) {
			struct qnx4fs_xb *xb2;
			xb2 = qnx4fs_xalloc(ip, &bp2, &xbnr);
			if (xb2 == NULL) {
				brelse(bp);
				goto fail_bfree;
			}
			xb->xb_next  = cpu_to_le32(xbnr);
			xb2->xb_prev = cpu_to_le32(QNX4FS_I(ip)->i_xbnr_last);
			QNX4FS_I(ip)->i_xbnr_last = xbnr;
			xb = xb2;
			ixt = 0;
			tmp = 1;
		}

		xb->xb_xts[ixt].start = cpu_to_le32(goal);
		xb->xb_xts[ixt].size = cpu_to_le32(1);
		xb->xb_nblocks = cpu_to_le32(tmp);

		if (bp2) {
			mark_buffer_dirty(bp2);
			brelse(bp2);
		} else {
			++xb->xb_nxt;
		}
		++QNX4FS_I(ip)->i_nxt;
	}

	++ip->i_blocks;
	QNX4FS_I(ip)->i_mmu_private += QNX4FS_BLOCK_SIZE;
	mark_inode_dirty(ip);
	if (bp) {
		mark_buffer_dirty(bp);
		brelse(bp);
	}

	return goal;

fail_bfree:
	/* we can't get xblk, so we have to free allocated block */
	if (res >= 0)
		qnx4fs_bfree(ip->i_sb, goal, 1);

fail:
	QNX4FS_WARN(" failed inode %ld\n", ip->i_ino);
	return 0;
}
