/*
 *  linux/fs/qnx4fs/inode.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  14 jan 2004 Dmitry Yakovlev - 2.6 kernel support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/pagemap.h>
#include <linux/fs.h>
#include "qnx4fs_fs.h"
#include <linux/buffer_head.h>

static int
qnx4fs_count_iblocks(struct inode *ip)
{
	struct buffer_head *bp;
	struct qnx4fs_xb *xb;
	unsigned long iblocks, xbnr;

	iblocks = (ip->i_size + QNX4FS_BLOCK_SIZE-1) >> QNX4FS_BLOCK_SIZE_BITS;

	xbnr = QNX4FS_I(ip)->i_xbnr0;
	while (xbnr) {
		bp = qnx4fs_sb_bread(ip->i_sb, xbnr - 1);
		if (bp == NULL) {
			QNX4FS_WARN(" can't read block %lu\n", xbnr);
			return 0;
		}
		xb = (struct qnx4fs_xb *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			brelse(bp);
			QNX4FS_WARN(" wrong xblk %lu signature\n", xbnr);
			return 0;
		}
		++iblocks;
		QNX4FS_I(ip)->i_xbnr_last = xbnr;
		xbnr = le32_to_cpu(xb->xb_next);
		brelse(bp);
	}
	ip->i_blocks = iblocks;
	QNX4FS_DEBUG(" inode %ld, blocks %d\n", ip->i_ino, (int)iblocks);
	return 1;
}

struct inode *
qnx4fs_iget(struct super_block *sb, unsigned long ino)
{
	struct buffer_head *bp;
	qnx4fs_dinode_t *qip;
	struct inode *ip;

	ip = iget_locked(sb, ino);
	if (!ip)
		return ERR_PTR(-ENOMEM);

	if (!(ip->i_state & I_NEW))
		return ip;

	bp = qnx4fs_sb_bread(ip->i_sb, ino2phys(ino));
	if (bp == NULL)
		goto fail;
	qip = (qnx4fs_dinode_t *)bp->b_data + ino2slot(ino);

	if (!(qip->i.status & QNX4FS_FILE_USED) ||
		(qip->i.status & QNX4FS_FILE_LINK)) {
		QNX4FS_WARN(" martian inode status %02X, ", qip->i.status);
		goto fail;
	}

	ip->i_mode	= le16_to_cpu(qip->i.mode);
	ip->i_uid	= (uid_t) le16_to_cpu(qip->i.uid);
	ip->i_gid	= (gid_t) le16_to_cpu(qip->i.gid);
	ip->i_nlink	= le16_to_cpu(qip->i.nlink);
	ip->i_size	= le32_to_cpu(qip->i.size);
	ip->i_mtime.tv_sec  = le32_to_cpu(qip->i.mtime);
	ip->i_mtime.tv_nsec = 0;
	ip->i_atime.tv_sec  = le32_to_cpu(qip->i.atime);
	ip->i_atime.tv_nsec = 0;
	ip->i_ctime.tv_sec  = le32_to_cpu(qip->i.ctime);
	ip->i_ctime.tv_nsec = 0;

	QNX4FS_I(ip)->i_nxt		= le16_to_cpu(qip->i.nxt);
	QNX4FS_I(ip)->i_xbnr		= QNX4FS_I(ip)->i_xbnr0
					= le32_to_cpu(qip->i.xb);
	QNX4FS_I(ip)->i_phys		= QNX4FS_I(ip)->i_xt0.start
					= le32_to_cpu(qip->i.xt0.start);
	QNX4FS_I(ip)->i_offset_end	= QNX4FS_I(ip)->i_xt0.size
					= le32_to_cpu(qip->i.xt0.size);
	QNX4FS_I(ip)->i_offset		= 0;
	QNX4FS_I(ip)->i_xtnr		= -1;

	QNX4FS_I(ip)->i_linked		= qip->i.status & QNX4FS_FILE_INODE;
	QNX4FS_I(ip)->i_mmu_private	= ip->i_size;

	QNX4FS_DEBUG(" inode %ld, start %ld\n",
		 ip->i_ino, QNX4FS_I(ip)->i_phys);
	brelse(bp);

	if (S_ISREG(ip->i_mode)) {
		ip->i_op = &qnx4fs_file_inode_ops;
		ip->i_fop = &qnx4fs_file_ops;
		ip->i_mapping->a_ops = &qnx4fs_address_space_ops;
	} else if (S_ISDIR(ip->i_mode)) {
		ip->i_op = &qnx4fs_dir_inode_ops;
		ip->i_fop = &qnx4fs_dir_ops;
		ip->i_mapping->a_ops = &qnx4fs_address_space_ops;
	} else if (S_ISLNK(ip->i_mode)) {
		ip->i_op = &page_symlink_inode_operations;
		ip->i_mapping->a_ops = &qnx4fs_address_space_ops;
	} else {
		goto fail;
	}
	if (!qnx4fs_count_iblocks(ip))
		goto fail;

	unlock_new_inode(ip);
	return ip;
fail:
	iget_failed(ip);
	QNX4FS_WARN(" fail ino %d\n", (int)ino);
	return ERR_PTR(-EIO);
}

int
qnx4fs_write_inode(struct inode *ip, int wait)
{
	struct buffer_head *bp;
	qnx4fs_dinode_t *qip;
	ino_t ino;

	/* mutex_lock(&ip->i_mutex); */
	mutex_lock(&QNX4FS_SB(ip->i_sb)->s_slock);

	ino = ip->i_ino;
	QNX4FS_DEBUG(" inode %ld, wait %d\n", ip->i_ino, wait);

	bp = qnx4fs_sb_bread(ip->i_sb, ino2phys(ino));
	if (bp == NULL) {
		mutex_unlock(&QNX4FS_SB(ip->i_sb)->s_slock);
		return -EIO;
	}

	qip = (qnx4fs_dinode_t *)bp->b_data + ino2slot(ino);

	if (!(qip->i.status & QNX4FS_FILE_USED) ||
	(qip->i.status & QNX4FS_FILE_LINK)) {
		if (qip->i.status)
			QNX4FS_WARN(" inode %ld, status %x\n",
			ip->i_ino, qip->i.status);
		brelse(bp);
		mutex_unlock(&QNX4FS_SB(ip->i_sb)->s_slock);
		return -EIO;
	}

	qip->i.size	 = cpu_to_le32(ip->i_size);
	qip->i.mode	 = cpu_to_le16(ip->i_mode);
	qip->i.uid	 = cpu_to_le16(ip->i_uid);
	qip->i.gid	 = cpu_to_le16(ip->i_gid);
	qip->i.nlink	 = cpu_to_le16(ip->i_nlink);
	qip->i.atime	 = cpu_to_le32(ip->i_atime.tv_sec);
	qip->i.ctime	 = cpu_to_le32(ip->i_ctime.tv_sec);
	qip->i.mtime	 = cpu_to_le32(ip->i_mtime.tv_sec);
	qip->i.xt0.start = cpu_to_le32(QNX4FS_I(ip)->i_xt0.start);
	qip->i.xt0.size	 = cpu_to_le32(QNX4FS_I(ip)->i_xt0.size);
	qip->i.xb	 = cpu_to_le32(QNX4FS_I(ip)->i_xbnr0);
	qip->i.nxt	 = cpu_to_le16(QNX4FS_I(ip)->i_nxt);

	mutex_unlock(&QNX4FS_SB(ip->i_sb)->s_slock);
	/* mutex_unlock(&ip->i_mutex); */
	mark_buffer_dirty(bp);
	brelse(bp);
	return 0;
}

void
qnx4fs_delete_inode(struct inode *ip)
{
	ino_t ino = ip->i_ino;
	int linked = QNX4FS_I(ip)->i_linked;

	QNX4FS_DEBUG(" inode %ld\n", ip->i_ino);
	truncate_inode_pages(&ip->i_data, 0);
	ip->i_size = 0;
	qnx4fs_truncate(ip);
	if (linked)
		qnx4fs_difree(ip, ino);
	clear_inode(ip);
}
