/*
 *  linux/fs/qnx4fs/namei.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  01 oct 2002 Dmitry Yakovlev - some fixes.
 *  14 jan 2004 Dmitry Yakovlev - 2.6 kernel support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/smp_lock.h>
#include "qnx4fs_fs.h"
#include <linux/buffer_head.h>

static inline void
qnx4fs_link_dirent(qnx4fs_dirent_t *qdp, ino_t ino)
{
	qdp->l.block  = cpu_to_le32(ino2phys(ino) + 1);
	qdp->l.index  = ino2slot(ino);
	qdp->l.status = QNX4FS_FILE_LINK;
}

static inline void
qnx4fs_change_ino(struct inode *ip, ino_t new_ino, int linked)
{
	/* FIXME: this maybe incorrect */
	mutex_lock(&QNX4FS_SB(ip->i_sb)->s_slock);
	remove_inode_hash(ip);
	ip->i_ino = new_ino;
	insert_inode_hash(ip);
	if (linked)
		QNX4FS_I(ip)->i_linked = 1;
	mutex_unlock(&QNX4FS_SB(ip->i_sb)->s_slock);
	mark_inode_dirty(ip);
}


static qnx4fs_dirent_t *
qnx4fs_find_dirent(struct inode *dip, struct dentry *dp,
		    struct buffer_head **bpp, ino_t *inop)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	long block, nblocks;
	const char *name  = dp->d_name.name;
	int slot, namelen = dp->d_name.len;
	ino_t ino;

	QNX4FS_DEBUG(" inode %ld, name <%s>\n", dip->i_ino, dp->d_name.name);
	nblocks = qnx4fs_nblocks(dip);
	for (block = 0; block < nblocks; block++) {
		phys = qnx4fs_map_block(dip, block);
		if (!phys) {
			QNX4FS_WARN(" can't map block %ld\n", block);
			return NULL;
		}
		bp = qnx4fs_sb_bread(dip->i_sb, phys - 1);
		if (bp == NULL) {
			QNX4FS_WARN(" can't read block %lu\n", phys);
			return NULL;
		}

		qdp = (qnx4fs_dirent_t *)bp->b_data;
		for (slot = 0; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {
			if (qdp->l.status & QNX4FS_FILE_LINK) {
				if (strncmp(qdp->l.name, name,
					    QNX4FS_LONG_NAME_MAX) == 0) {
					ino = link2ino(qdp);
					goto found;
				}
			} else if ((qdp->i.status & QNX4FS_FILE_USED)
				   && namelen <= QNX4FS_SHORT_NAME_MAX
				   && strncmp(qdp->i.name, name,
					       QNX4FS_SHORT_NAME_MAX) == 0) {
				ino = direct2ino(phys, slot);
				goto found;
			}
		}
		brelse(bp);
	}

	return NULL;
found:
	if (ino == QNX4FS_INODES_INO ||
		ino == QNX4FS_SB(dip->i_sb)->s_bitmap_ino) {
		brelse(bp);
		return NULL;
	}
	*inop = ino;
	*bpp = bp;
	return qdp;
}

static int
qnx4fs_is_dir_empty(struct inode *dip)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	long block, nblocks;
	unsigned long phys;
	int slot = 2;

	nblocks = qnx4fs_nblocks(dip);
	for (block = 0; block < nblocks; block++) {
		phys = qnx4fs_map_block(dip, block);
		if (!phys)
			return 0;
		bp = qnx4fs_sb_bread(dip->i_sb, phys - 1);
		if (bp == NULL)
			return 0;
		qdp = (qnx4fs_dirent_t *)bp->b_data + slot;
		for (; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {
			if (!qdp->i.status)
				continue;
			if (qdp->i.status &
			(QNX4FS_FILE_USED | QNX4FS_FILE_LINK)) {
				brelse(bp);
				return 0;
			}
			if (qdp->i.status
				&& (dip->i_ino == QNX4FS_ROOT_INO)) {
				brelse(bp);
				return 0;
			}
			if (qdp->i.status)
				QNX4FS_WARN(" wrong %ld status %02X\n",
				direct2ino(phys, slot), qdp->i.status);
		}
		slot = 0;
		brelse(bp);
	}

	return 1;
}

static int
qnx4fs_make_empty_dir(struct inode *dip, struct inode *ip)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;

	bp = qnx4fs_iexpand_special(ip, 4);

	if (bp == NULL)
		return -EIO; /* ? -ENOSPC */

	qdp = (qnx4fs_dirent_t *)bp->b_data;

	memcpy(qdp->d.signature, ".\0I\3QNX\0", 8);
	/* This makes chkfsys happy... */
	qdp->d.block = cpu_to_le32(QNX4FS_I(ip)->i_xt0.size);
	qnx4fs_link_dirent(qdp++, ip->i_ino);
	memcpy(qdp->l.name, "..\0", 3);

	qnx4fs_link_dirent(qdp, dip->i_ino);
	qnx4fs_inc_nlink(ip);
	mark_buffer_dirty(bp);
	brelse(bp);

	return 0;
}

static int
qnx4fs_fill_dirent(qnx4fs_dirent_t *qdp, struct buffer_head *bp,
		struct dentry *dp, struct inode *ip)
{
	struct inode *dip = dp->d_parent->d_inode;
	int tail;

	memcpy(qdp->i.name, dp->d_name.name, dp->d_name.len);

	if (QNX4FS_I(ip)->i_linked) {
		tail = QNX4FS_LONG_NAME_MAX - dp->d_name.len;
		qnx4fs_link_dirent(qdp, ip->i_ino);
	} else {
		tail = QNX4FS_SHORT_NAME_MAX - dp->d_name.len;
		qdp->i.status = QNX4FS_FILE_USED;
	}

	if(tail > 0)
		memset(qdp->i.name + dp->d_name.len, 0, tail);

	dip->i_mtime = CURRENT_TIME;
	dip->i_ctime = CURRENT_TIME;
	mark_inode_dirty(dip);
	mark_buffer_dirty(bp);
	brelse(bp);

	return 0;
}

static inline int
qnx4fs_link_inode(struct inode *ip, ino_t new_ino)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	ino_t ino;

	ino = ip->i_ino;
	bp = qnx4fs_sb_bread(ip->i_sb, ino2phys(ino));
	if (bp == NULL)
		return -EIO;
	qdp = (qnx4fs_dirent_t *)bp->b_data + ino2slot(ino);
	qnx4fs_link_dirent(qdp, new_ino);
	mark_buffer_dirty(bp);
	brelse(bp);
	qnx4fs_change_ino(ip, new_ino, 1);
	return 0;
}

static int
qnx4fs_instantiate_dirent(qnx4fs_dirent_t *qdp, struct buffer_head *bp,
			   struct dentry *dp, struct inode *ip)
{
	int err;

	err = qnx4fs_fill_dirent(qdp, bp, dp, ip);
	if (err) {
		/* inode_dec_link_count(ip); */
		ip->i_nlink = 0;
		mark_inode_dirty(ip);
		iput(ip);
	} else {
		mark_inode_dirty(ip);
		d_instantiate(dp, ip);
	}
	return err;
}

static struct dentry *
qnx4fs_lookup(struct inode *dip, struct dentry *dp,
		struct nameidata *nd)

{
	struct inode *ip = NULL;
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	ino_t ino;

	QNX4FS_DEBUG(" inode %ld, name <%s>\n", dip->i_ino, dp->d_name.name);
	if (dp->d_name.len > QNX4FS_LONG_NAME_MAX)
		return ERR_PTR(-ENAMETOOLONG);

	/* mutex_lock(&QNX4FS_SB(dip->i_sb)->s_slock); */
	qdp = qnx4fs_find_dirent(dip, dp, &bp, &ino);
	if (qdp) {
		brelse(bp);
		ip = qnx4fs_iget(dip->i_sb, ino);
		if (IS_ERR(ip)) {
			/* mutex_unlock(&QNX4FS_SB(dip->i_sb)->s_slock); */
			return ERR_CAST(ip);
		}
	}

	d_add(dp, ip);
	/* mutex_unlock(&QNX4FS_SB(dip->i_sb)->s_slock); */

	return NULL;
}

static int
qnx4fs_create(struct inode *dip, struct dentry *dp, int mode,
		struct nameidata *nd)

{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	struct inode *ip;
	int err;

	QNX4FS_DEBUG(" inode %ld, name <%s>\n", dip->i_ino, dp->d_name.name);
	bp = qnx4fs_inew(dip, dp, &ip, &qdp);
	if (IS_ERR(bp))
		return PTR_ERR(bp);

	ip->i_mode		= mode;
	ip->i_op		= &qnx4fs_file_inode_ops;
	ip->i_fop		= &qnx4fs_file_ops;
	ip->i_mapping->a_ops	= &qnx4fs_address_space_ops;

	insert_inode_hash(ip);
	err = qnx4fs_instantiate_dirent(qdp, bp, dp, ip);
	return err;
}

static int
qnx4fs_link(struct dentry *old_dp, struct inode *dip, struct dentry *dp)
{
	struct inode *ip = old_dp->d_inode;
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	ino_t ino;
	int err;

	if (S_ISDIR(ip->i_mode))
		return -EPERM;

	if (ip->i_nlink >= QNX4FS_LINK_MAX)
		return -EMLINK;

	if (!QNX4FS_I(ip)->i_linked) {
		err = qnx4fs_dialloc(ip->i_sb, &ino);
		if (err)
			goto out;
		err = qnx4fs_link_inode(ip, ino);
		if (err) {
			qnx4fs_difree(ip, ino);
			goto out;
		}
	}

	bp = qnx4fs_get_slot(dip, &ino, &qdp, QNX4FS_DIR_EXPAND_DELTA);

	if (bp == NULL) {
		err = -ENOSPC;
		goto out;
	}

	err = qnx4fs_fill_dirent(qdp, bp, dp, ip);

	if (!err) {
		ip->i_ctime = CURRENT_TIME;
		qnx4fs_inc_nlink(ip);
		atomic_inc(&ip->i_count);
		d_instantiate(dp, ip);
	}
out:
	return err;
}

static inline int
qnx4fs_move_to_inodes(struct inode *ip)
{
	ino_t ino;
	int err;
	err = qnx4fs_dialloc(ip->i_sb, &ino);
	if (err)
		return err;
	qnx4fs_change_ino(ip, ino, 1);
	return err;
}

static int
qnx4fs_unlink(struct inode *dip, struct dentry *dp)
{
	struct inode *ip = dp->d_inode;
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	ino_t ino;
	int err = 0;

	QNX4FS_DEBUG(" inode %ld, name <%s>\n", dip->i_ino, dp->d_name.name);
	qdp = qnx4fs_find_dirent(dip, dp, &bp, &ino);

	if (qdp == NULL)
		return -ENOENT;

	qnx4fs_mark_slot(qdp, bp, 0);

	/* if (QNX4FS_I(ip)->i_linked == 0 && atomic_read(&ip->i_count) > 1)
	 * I dont know how long it will be in the system
	 * until it will be desroyed
	 */
	if (QNX4FS_I(ip)->i_linked == 0)
		err = qnx4fs_move_to_inodes(ip);

/*	QNX4FS_DEBUG(" ino %d -> %d, name <%s> i_count %d\n",
		(int)ip->i_ino, (int)ino, dp->d_name.name,
		(int)atomic_read(&ip->i_count));
*/

	ip->i_ctime = dip->i_ctime;
	qnx4fs_dec_nlink(ip);

	/* we shold not return -ENOSPC here */
	return 0;
}

static int
qnx4fs_symlink(struct inode *dip, struct dentry *dp, const char *symname)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	struct inode *ip;
	int err, len = strlen(symname);

	if (len > QNX4FS_BLOCK_SIZE)
		return -ENAMETOOLONG;

	bp = qnx4fs_inew(dip, dp, &ip, &qdp);
	if (IS_ERR(bp))
		return PTR_ERR(bp);

	ip->i_mode		= S_IFLNK | S_IRWXUGO;
	ip->i_op		= &page_symlink_inode_operations;
	ip->i_mapping->a_ops	= &qnx4fs_address_space_ops;

	insert_inode_hash(ip);
	err = page_symlink(ip, symname, len + 1);
	if (err) {
		brelse(bp);
		qnx4fs_dec_nlink(ip);
		iput(ip);
		return err;
	}
	err = qnx4fs_instantiate_dirent(qdp, bp, dp, ip);
	return err;
}

static int
qnx4fs_mkdir(struct inode *dip, struct dentry *dp, int mode)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	struct inode *ip;
	int err;

	QNX4FS_DEBUG(" inode %ld, name <%s>\n", dip->i_ino, dp->d_name.name);

	if (dip->i_nlink >= QNX4FS_LINK_MAX)
		return -EMLINK;

	qnx4fs_inc_nlink(dip);
	/* inode_inc_link_count(dip); */
	bp = qnx4fs_inew(dip, dp, &ip, &qdp);

	if (IS_ERR(bp)) {
		err = PTR_ERR(bp);
		goto fail;
	}

	ip->i_mode		= S_IFDIR | mode;
	ip->i_op		= &qnx4fs_dir_inode_ops;
	ip->i_fop		= &qnx4fs_dir_ops;
	ip->i_mapping->a_ops	= &qnx4fs_address_space_ops;

	insert_inode_hash(ip);
	err = qnx4fs_make_empty_dir(dip, ip);
	if (err) {
		brelse(bp);
		qnx4fs_dec_nlink(ip);
		iput(ip);
		goto fail;
	}

	err = qnx4fs_instantiate_dirent(qdp, bp, dp, ip);
	if (!err)
		return 0;

fail:
	qnx4fs_dec_nlink(dip);
	return err;
}

static int
qnx4fs_rmdir(struct inode *dip, struct dentry *dp)
{
	struct inode *ip = dp->d_inode;
	int err = -ENOTEMPTY;

	if (qnx4fs_is_dir_empty(ip)) {
		err = qnx4fs_unlink(dip, dp);
		if (!err) {
			qnx4fs_dec_nlink(dip);
			qnx4fs_dec_nlink(ip);
		}
	}
	return err;
}

static int
qnx4fs_rename_helper(struct inode *ip, struct dentry *dp, ino_t *inop)
{
	int err;

	if (QNX4FS_I(ip)->i_linked) {
		*inop = ip->i_ino;
	} else {
		if (dp->d_name.len > QNX4FS_SHORT_NAME_MAX) {
			err = qnx4fs_dialloc(ip->i_sb, inop);
			if (err)
				return err;

			qnx4fs_change_ino(ip, *inop, 1);
			return 0;
		}
		qnx4fs_change_ino(ip, *inop, 0);
	}

	return 0;
}

static int
qnx4fs_rename(struct inode *old_dip, struct dentry *old_dp,
	       struct inode *new_dip, struct dentry *new_dp)
{
	struct inode *old_ip = old_dp->d_inode;
	struct inode *new_ip = new_dp->d_inode;
	struct buffer_head *old_bp, *new_bp, *dir_bp = NULL;
	qnx4fs_dirent_t *old_qdp, *new_qdp, *dir_qdp = NULL;
	unsigned long phys;
	ino_t ino;
	int err;

	old_qdp = qnx4fs_find_dirent(old_dip, old_dp, &old_bp, &ino);

	if (old_qdp == NULL)
		return -ENOENT;

	if (S_ISDIR(old_ip->i_mode)) {
		err = -EIO;
		phys = qnx4fs_map_block(old_ip, 0);
		if (!phys)
			goto out_old_bp;
		dir_bp = qnx4fs_sb_bread(old_ip->i_sb, phys - 1);
		if (dir_bp == NULL)
			goto out_old_bp;

		dir_qdp = (qnx4fs_dirent_t *)dir_bp->b_data;
		QNX4FS_DEBUG(" QNX4FS rename dir old_dip %d new_dip %d," \
			"old_ino %d, linked %d\n",
			(int)old_dip->i_ino, (int)new_dip->i_ino,
			(int)old_ip->i_ino, (int)QNX4FS_I(old_ip)->i_linked);
	}

	if (new_ip == NULL) {
		if (dir_qdp && new_dip->i_nlink >= QNX4FS_LINK_MAX) {
			err = -EMLINK;
			goto out_old_bp;
		}
		qnx4fs_inc_nlink(old_ip);
		new_bp = qnx4fs_get_slot(new_dip, &ino, &new_qdp,
					 QNX4FS_DIR_EXPAND_DELTA);
		if (IS_ERR(new_bp)) {
			qnx4fs_dec_nlink(old_ip);
			err = PTR_ERR(new_bp);
			goto out_old_bp;
		}
		err = qnx4fs_rename_helper(old_ip, new_dp, &ino);
		if (err) {
			qnx4fs_dec_nlink(old_ip);
			brelse(new_bp);
			goto out_old_bp;
		}
		err = qnx4fs_fill_dirent(new_qdp, new_bp, new_dp, old_ip);
		if (err) {
			qnx4fs_dec_nlink(old_ip);
			goto out_old_bp;
		}
		if (dir_qdp)
			qnx4fs_inc_nlink(new_dip);
	} else {

		if (dir_qdp && !qnx4fs_is_dir_empty(new_ip)) {
			err = -ENOTEMPTY;
			goto out_old_bp;
		}
		new_qdp = qnx4fs_find_dirent(new_dip, new_dp, &new_bp, &ino);
		if (new_qdp == NULL) {
			err = -ENOENT;
			goto out_old_bp;
		}
		/* if the new file exists we move its inode to /.inodes
		 * (unlink likewise and for the same reason)
		 */
		/* if (QNX4FS_I(new_ip)->i_linked == 0
		 * && atomic_read(&new_ip->i_count) > 1)
		 */
		if (QNX4FS_I(new_ip)->i_linked == 0) {
			err = qnx4fs_move_to_inodes(new_ip);
			if (err) {
				brelse(new_bp);
				goto out_old_bp;
			}
		}
		qnx4fs_inc_nlink(old_ip);
		err = qnx4fs_rename_helper(old_ip, new_dp, &ino);
		if (err) {
			qnx4fs_dec_nlink(old_ip);
			brelse(new_bp);
			goto out_old_bp;
		}
		err = qnx4fs_fill_dirent(new_qdp, new_bp, new_dp, old_ip);
		if (err) {
			qnx4fs_dec_nlink(old_ip);
			goto out_old_bp;
		}
		new_ip->i_ctime = CURRENT_TIME;
		if (dir_qdp)
			/* drop_nlink(new_ip); */
			qnx4fs_dec_nlink(new_ip);
		qnx4fs_dec_nlink(new_ip);
	}
	qnx4fs_mark_slot(old_qdp, old_bp, 0);
	qnx4fs_dec_nlink(old_ip);

	if (dir_qdp) {
		qnx4fs_dec_nlink(old_dip);
		qnx4fs_link_dirent(dir_qdp, ino); /* dot entry */
		dir_qdp++;
		qnx4fs_link_dirent(dir_qdp, new_dip->i_ino); /* dot dot entry */
		mark_buffer_dirty(dir_bp);
		QNX4FS_DEBUG(" QNX4FS rename dir . %d .. %d\n",
			 (int)ino, (int)new_dip->i_ino);
		brelse(dir_bp);
	}

	return 0;

out_old_bp:
	brelse(old_bp);
	if (dir_qdp)
		brelse(dir_bp);
	return err;
}

struct inode_operations qnx4fs_dir_inode_ops = {
	.create		= qnx4fs_create,
	.lookup		= qnx4fs_lookup,
	.link		= qnx4fs_link,
	.unlink		= qnx4fs_unlink,
	.symlink	= qnx4fs_symlink,
	.mkdir		= qnx4fs_mkdir,
	.rmdir		= qnx4fs_rmdir,
	.rename		= qnx4fs_rename,
};
