#ifndef __QNX4FS_FS_H
#define __QNX4FS_FS_H


#define QNX4FS_ROOT_INO		8
#define QNX4FS_INODES_INO	9
#define QNX4FS_BOOT_INO		10
#define QNX4FS_ALTBOOT_INO	11

#define QNX4FS_LINK_MAX		65535
#define QNX4FS_SUPER_MAGIC	0x34584e51 /* little endian "QNX4" */

#define QNX4FS_BLOCK_SIZE	512
#define QNX4FS_BLOCK_SIZE_BITS	9
#define QNX4FS_DIRENT_SIZE	64
#define QNX4FS_DIRENT_SIZE_BITS	6
#define QNX4FS_INODES_PER_BLOCK	8
#define QNX4FS_INODES_PER_BLOCK_BITS	3
#define QNX4FS_DIRENTS_PER_BLOCK	QNX4FS_INODES_PER_BLOCK
#define QNX4FS_DIRENTS_PER_BLOCK_BITS	QNX4FS_INODES_PER_BLOCK_BITS

#define QNX4FS_BOOT_BLOCK	1L
#define QNX4FS_ROOT_BLOCK	2L
#define QNX4FS_BITMAP_BLOCK	3L

#define QNX4FS_MAX_XTS_PER_XBLK	60
#define QNX4FS_LONG_NAME_MAX	48
#define QNX4FS_SHORT_NAME_MAX	16

#define QNX4FS_FILE_USED	0x01
#define QNX4FS_FILE_MODIFIED	0x02
#define QNX4FS_FILE_BUSY	0x04
#define QNX4FS_FILE_LINK	0x08
#define QNX4FS_FILE_INODE	0x10
#define QNX4FS_FSYS_CLEAN	0x20
#define QNX4FS_FILE_GROWN	0x40

#define QNX4FS_DIR_EXPAND_DELTA	4
#define QNX4FS_INODES_EXPAND_DELTA	16

/* These are the on-disk formats and in little endian */

struct qnx4fs_xt {
	__le32 start;
	__le32 size;
};

struct qnx4fs_xb {
	__le32			xb_next;
	__le32			xb_prev;
	u8			xb_nxt;
	char			xb_spare[3];
	__le32			xb_nblocks;
	struct qnx4fs_xt	xb_xts[QNX4FS_MAX_XTS_PER_XBLK];
	char			xb_signature[8];
	struct qnx4fs_xt	xb_xt0;
};

struct qnx4fs_dinode {
	char			name[QNX4FS_SHORT_NAME_MAX];
	__le32			size;
	struct qnx4fs_xt	xt0;
	__le32			xb;
	__le32			ftime;
	__le32			mtime;
	__le32			atime;
	__le32			ctime;
	__le16			nxt;
	__le16			mode;
	__le16			uid;
	__le16			gid;
	__le16			nlink;
	u8			zero[4];
	u8			type;
	u8			status;
};

struct qnx4fs_dlink {
	char	name[QNX4FS_LONG_NAME_MAX];
	__le32	block;
	u8	index;
	char	spare[10];
	u8	status;
};

struct qnx4fs_dotentry {
	char	signature[8];
	__le32	block;
};
typedef union qnx4fs_dirent  {
	struct qnx4fs_dinode i;
	struct qnx4fs_dlink l;
	struct qnx4fs_dotentry d;
} qnx4fs_dirent_t;

typedef qnx4fs_dirent_t qnx4fs_dinode_t;

/* #define  QNX4DBG */

#ifdef QNX4DBG
#define QNX4FS_DEBUG(f, a...)	printk(KERN_INFO "QNX4FS_DEBUG %s: %s:" f, \
		__FILE__, __func__, ## a)
#else
#define QNX4FS_DEBUG(f, a...)	/**/
#endif

#define QNX4FS_WARN(f, a...)	printk(KERN_WARNING "QNX4FS_WARN %s: " f, \
		__func__, ## a)

#ifdef __KERNEL__

/* qnx4fs filesystem in-core inode info
 * all of these are in cpu byte order  */

struct qnx4fs_xt_i {
	u32 start;
	u32 size;
};

struct qnx4fs_inode_info {
	int			i_nxt;
	int			i_xtnr;
	struct qnx4fs_xt_i	i_xt0;
	unsigned long		i_xbnr;
	unsigned long		i_xbnr0;
	unsigned long		i_xbnr_last;
	long			i_offset;
	long			i_offset_end;
	unsigned long		i_phys;
	loff_t			i_mmu_private;
	int			i_linked;
	struct inode 		vfs_inode;
};
/* qnx4fs in-core superblock data */
struct qnx4fs_sb_info {
	ino_t			s_bitmap_ino;
	unsigned long		s_bitmap_blocks;
	struct mutex	 	s_bmlock;	/* Protects bitmap access */
	struct mutex	 	s_iflock;	/* Protects /.inodes files */
	struct mutex		s_slock;	/* Protects the super block */
	unsigned long		s_blocks;
	struct inode		*s_inodes;
	unsigned long		s_last_iblock;
};

/* bitmap.c */
extern void qnx4fs_bfree(struct super_block *, unsigned long, unsigned long);
extern int qnx4fs_balloc(struct super_block *, unsigned long *);
extern unsigned long qnx4fs_count_free_blocks(struct super_block *);

/* common.c */

static inline ino_t
direct2ino(unsigned long phys, int slot)
{
	return ((phys-1) << QNX4FS_INODES_PER_BLOCK_BITS) + slot;
}

static inline void
qnx4fs_inc_nlink(struct inode *ip)
{
	ip->i_nlink++;
	mark_inode_dirty(ip);
}

static inline void
qnx4fs_dec_nlink(struct inode *ip)
{
	ip->i_nlink--;
	mark_inode_dirty(ip);
}

static inline unsigned long
ino2phys(ino_t ino)
{
	return ino >> QNX4FS_INODES_PER_BLOCK_BITS;
}

static inline int
ino2slot(ino_t ino)
{
	return ino & (QNX4FS_INODES_PER_BLOCK-1);
}

static inline ino_t
link2ino(qnx4fs_dirent_t *qdp)
{
	return direct2ino(le32_to_cpu(qdp->l.block), qdp->l.index);
}

static inline long
qnx4fs_nblocks(struct inode *ip)
{
	return (ip->i_size + QNX4FS_BLOCK_SIZE-1) >> QNX4FS_BLOCK_SIZE_BITS;
}

extern struct buffer_head *qnx4fs_get_slot(struct inode *, ino_t *,
					    qnx4fs_dirent_t **, int);
extern void qnx4fs_mark_slot(qnx4fs_dirent_t *, struct buffer_head *, int);

/* dir.c */
extern const struct file_operations qnx4fs_dir_ops;

/* file.c */
extern const struct file_operations qnx4fs_file_ops;
extern struct inode_operations qnx4fs_file_inode_ops;
extern struct address_space_operations qnx4fs_address_space_ops;

extern int qnx4fs_get_block(struct inode *ip, sector_t block,
		  struct buffer_head *bh_result, int create);

extern unsigned long qnx4fs_map_block(struct inode *, long);

/* ialloc.c */
extern int qnx4fs_dialloc(struct super_block *, ino_t *);
extern void qnx4fs_difree(struct inode*, ino_t);
extern struct buffer_head *qnx4fs_inew(struct inode *, struct dentry *,
					struct inode **, qnx4fs_dirent_t **);
extern unsigned long qnx4fs_iexpand(struct inode *);
extern struct buffer_head *qnx4fs_iexpand_special(struct inode *, int);

/* inode.c */
extern struct inode *qnx4fs_iget(struct super_block *, unsigned long);
extern void qnx4fs_put_inode(struct inode *);
extern int qnx4fs_write_inode(struct inode *, int);
extern void qnx4fs_delete_inode(struct inode *);

/* namei.c */
extern struct inode_operations qnx4fs_dir_inode_ops;

/* super.c */
extern struct super_operations qnx4fs_super_ops;

/* truncate.c */
extern void qnx4fs_truncate(struct inode *);

static inline struct qnx4fs_sb_info *
QNX4FS_SB(struct super_block *sb)
{
	return sb->s_fs_info;
}

static inline struct qnx4fs_inode_info *
QNX4FS_I(struct inode *inode)
{
	return container_of(inode, struct qnx4fs_inode_info, vfs_inode);
}

extern struct buffer_head*
qnx4fs_sb_bread(struct super_block *sb, unsigned long block);

#endif /* __KERNEL__ */
#endif /* __QNX4FS_FS_H */
