/*
 *  linux/fs/qnx4fs/super.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  09 oct 2002 Dmitry Yakovlev - some fixes.
 *  14 jan 2004 Dmitry Yakovlev - 2.6 kernel support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/buffer_head.h>
#include <linux/vfs.h>
#include "qnx4fs_fs.h"

static int
qnx4fs_fill_super(struct super_block *sbp, void *dp, int silent)
{
	struct buffer_head *bp;
	struct inode *ip, *root;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	struct qnx4fs_sb_info *qsb;

	QNX4FS_DEBUG("\n");
	qsb = kzalloc(sizeof(struct qnx4fs_sb_info), GFP_KERNEL);
	if (!qsb)
		return -ENOMEM;

	sbp->s_fs_info = qsb;
	mutex_init(&qsb->s_bmlock);
	mutex_init(&qsb->s_iflock);
	mutex_init(&qsb->s_slock);

	if (!sb_set_blocksize(sbp, QNX4FS_BLOCK_SIZE)) {
		printk(KERN_ERR "qnx4fs: blocksize too small for device.\n");
		goto out;
	}

	bp = sb_bread(sbp, QNX4FS_ROOT_BLOCK - 1);
	if (bp == NULL) {
		printk(KERN_WARNING "qnx4fs: unable to read root block\n");
		goto out;
	}

	qdp = (qnx4fs_dirent_t *)bp->b_data;

	if (memcmp(qdp->i.name, "/", 2)) {
		if (!silent)
			printk(KERN_WARNING "qnx4fs: no root dir\n");
		brelse(bp);
		goto out;
	}

	phys = le32_to_cpu(qdp->i.xt0.start);
	QNX4FS_SB(sbp)->s_bitmap_ino = direct2ino(phys, 2);
	QNX4FS_SB(sbp)->s_bitmap_blocks = phys - QNX4FS_BITMAP_BLOCK;
	brelse(bp);

	bp = sb_bread(sbp, phys - 1);
	if (bp == NULL) {
		printk(KERN_WARNING "qnx4fs: unable to read root dir block\n");
		goto out;
	}
	qdp = (qnx4fs_dirent_t *)bp->b_data + 2;
	if (memcmp(qdp->i.name, ".bitmap\0", 8)) {
		if (!silent)
			printk(KERN_WARNING "qnx4fs: wrong '/.bitmap' entry\n");
		brelse(bp);
		goto out;
	}

	if (sbp->s_bdev && sbp->s_bdev->bd_inode)
		QNX4FS_SB(sbp)->s_blocks =
		sbp->s_bdev->bd_inode->i_size >> QNX4FS_BLOCK_SIZE_BITS;

	if (QNX4FS_SB(sbp)->s_blocks == 0) {
		if (!silent)
			printk(KERN_WARNING "qnx4fs: can't get perfect partition size\n");
		brelse(bp);
		QNX4FS_SB(sbp)->s_blocks = le32_to_cpu(qdp->i.size) << 3;
	}

	if ((le32_to_cpu(qdp->i.size) !=
		((QNX4FS_SB(sbp)->s_blocks) + 7) >> 3)) {
		if (!silent)
			printk(KERN_WARNING "qnx4fs: wrong /.bitmap size\n");
		brelse(bp);
		goto out;
	}

	brelse(bp);

	sbp->s_op = &qnx4fs_super_ops;
	sbp->s_magic = QNX4FS_SUPER_MAGIC;
	/* sbp->s_flags = MS_RDONLY; */

	/* we should iget() /.inodes inode before root dir inode */
	ip = qnx4fs_iget(sbp, QNX4FS_INODES_INO);
	if (IS_ERR(ip)) {
		printk(KERN_WARNING "qnx4fs: unable to get '/.inodes' inode\n");
		goto out;
	}
	QNX4FS_SB(sbp)->s_inodes = ip;

	root = qnx4fs_iget(sbp, QNX4FS_ROOT_INO);

	if (IS_ERR(root)) {
		printk(KERN_WARNING "qnx4fs: unable to get root inode\n");
		goto out_inodes;
	}

	sbp->s_root = d_alloc_root(root);
	if (sbp->s_root == NULL) {
		iput(root);
		printk(KERN_WARNING "qnx4fs: unable to get root dentry\n");
		goto out_inodes;
	}

	if (!(sbp->s_flags & MS_RDONLY))
		sbp->s_dirt = 1;

	printk(KERN_WARNING "qnx4fs mounted: blocks %lu, bitmap blocks %lu\n",
		QNX4FS_SB(sbp)->s_blocks, QNX4FS_SB(sbp)->s_bitmap_blocks);
	return 0;
out_inodes:
	iput(QNX4FS_SB(sbp)->s_inodes);
out:
	kfree(qsb);
	sbp->s_fs_info = NULL;
	return -EINVAL;
}

static void
qnx4fs_put_super(struct super_block *sb)
{
	struct qnx4fs_sb_info *qsb = QNX4FS_SB(sb);
	if (qsb) {
		iput(qsb->s_inodes);
		kfree(qsb);
	}
	sb->s_fs_info = NULL;
}

static void
qnx4fs_write_super(struct super_block *sbp)
{
	mutex_lock(&QNX4FS_SB(sbp)->s_slock);
	sbp->s_dirt = 0;
	mutex_unlock(&QNX4FS_SB(sbp)->s_slock);
}

static int
qnx4fs_remount_fs(struct super_block *sbp, int *flags, char *dp)
{
	return 0;
}

static int
qnx4fs_statfs(struct dentry *dentry, struct kstatfs *stat)
{
	struct super_block *sbp = dentry->d_sb;

	mutex_lock(&QNX4FS_SB(sbp)->s_slock);

	stat->f_type	= sbp->s_magic;
	stat->f_bsize	= sbp->s_blocksize;
	stat->f_blocks	= QNX4FS_SB(sbp)->s_blocks;
	stat->f_bfree	= stat->f_bavail
			= qnx4fs_count_free_blocks(sbp);
	stat->f_files	= 0;
	stat->f_ffree	= stat->f_bfree;
	stat->f_namelen	= QNX4FS_LONG_NAME_MAX;
	mutex_unlock(&QNX4FS_SB(sbp)->s_slock);

	return 0;
}

static struct inode *qnx4fs_alloc_inode(struct super_block *sb);

static void qnx4fs_destroy_inode(struct inode *inode);


struct super_operations qnx4fs_super_ops = {
	.alloc_inode	= qnx4fs_alloc_inode,
	.destroy_inode	= qnx4fs_destroy_inode,
	.write_inode    = qnx4fs_write_inode,
	.delete_inode	= qnx4fs_delete_inode,
	.put_super	= qnx4fs_put_super,
	.write_super	= qnx4fs_write_super,
	.statfs		= qnx4fs_statfs,
	.remount_fs	= qnx4fs_remount_fs,
};

static int
qnx4fs_get_sb(struct file_system_type *fs_type,
	int flags, const char *dev_name, void *data, struct vfsmount *mnt)
{
	return get_sb_bdev(fs_type, flags, dev_name, data,
		qnx4fs_fill_super, mnt);
}


static struct file_system_type qnx4fs_fs_type = {
	.owner		= THIS_MODULE,
	.name		= "qnx4fs",
	.get_sb		= qnx4fs_get_sb,
	.kill_sb	= kill_block_super,
	.fs_flags	= FS_REQUIRES_DEV,
};

static struct kmem_cache *qnx4fs_inode_cachep;

static void init_once(void *foo)
{
	struct qnx4fs_inode_info *ei = (struct qnx4fs_inode_info *) foo;
	inode_init_once(&ei->vfs_inode);
}

static int
init_inodecache(void)
{
	qnx4fs_inode_cachep =
	kmem_cache_create("qnx4fs_inode_cache",
			sizeof(struct qnx4fs_inode_info), 0,
			(SLAB_RECLAIM_ACCOUNT | SLAB_MEM_SPREAD),
			init_once);
	if (qnx4fs_inode_cachep == NULL)
		return -ENOMEM;
	return 0;
}

static struct inode *
qnx4fs_alloc_inode(struct super_block *sb)
{
	struct qnx4fs_inode_info *qi;
	qi = kmem_cache_alloc(qnx4fs_inode_cachep, GFP_KERNEL);
	if (!qi)
		return NULL;
	return &qi->vfs_inode;
}


static void
qnx4fs_destroy_inode(struct inode *inode)
{
	kmem_cache_free(qnx4fs_inode_cachep, QNX4FS_I(inode));
}

static void
destroy_inodecache(void)
{
	kmem_cache_destroy(qnx4fs_inode_cachep);
}

static int __init
qnx4fs_init(void)
{
	int err;

	err = init_inodecache();
	if (err)
		return err;
	err = register_filesystem(&qnx4fs_fs_type);
	if (err) {
		destroy_inodecache();
		return err;
	}
	printk(KERN_INFO "QNX4 filesystem registered.\n");
	return 0;
}

static void __exit
qnx4fs_cleanup(void)
{
	unregister_filesystem(&qnx4fs_fs_type);
	destroy_inodecache();
}

MODULE_AUTHOR("Serguei Tzukanov <tzukanov@narod.ru>");
MODULE_DESCRIPTION("QNX4 filesystem driver for Linux");
MODULE_LICENSE("GPL");

module_init(qnx4fs_init);
module_exit(qnx4fs_cleanup);
