/*
 *  linux/fs/qnx4fs/truncate.c
 *
 *  Copyright (C) 2001 Serguei Tzukanov
 *
 *  05 nov 2002 Dmitry Yakovlev - some fixes
 *  14 jan 2004 Dmitry Yakovlev - 2.6 kernel support
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
#include <linux/pagemap.h>
#include <linux/fs.h>
#include "qnx4fs_fs.h"
#include <linux/buffer_head.h>

static inline void
free_extent_le(struct super_block *sbp, struct qnx4fs_xt *xt)
{
	qnx4fs_bfree(sbp, le32_to_cpu(xt->start), le32_to_cpu(xt->size));
}

static inline void
shorten_extent_by(struct super_block *sbp, struct qnx4fs_xt *xt, int size)
{
	qnx4fs_bfree(sbp, le32_to_cpu(xt->start)
		+ le32_to_cpu(xt->size) - size, size);
	xt->size = cpu_to_le32(le32_to_cpu(xt->size) - size);
}

static inline void
truncate_xtnt_le(struct super_block *sbp, struct qnx4fs_xt *xtnt, int size)
{
	qnx4fs_bfree(sbp, le32_to_cpu(xtnt->start) + size,
		le32_to_cpu(xtnt->size) - size);
	xtnt->size = cpu_to_le32(size);
}

static inline struct buffer_head *
truncate_xblk(struct super_block *sbp, unsigned long nxblk, int nblocks) {
	struct buffer_head *bp;
	struct qnx4fs_xb *xblk;
	struct qnx4fs_xt *xtnt;
	int blocks, num_xtnts, xtnts;

	bp = sb_bread(sbp, nxblk - 1);
	if (bp == NULL) {
		QNX4FS_WARN("unable to finish truncate xblk %lu \n", nxblk);
		return 0;
	}

	xblk = (struct qnx4fs_xb *)bp->b_data;
	if (memcmp(xblk->xb_signature, "IamXblk\0", 8)) {
		QNX4FS_WARN(" wrong xblk %lu signature \n", nxblk);
		brelse(bp);
		return 0;
	}

	num_xtnts = xblk->xb_nxt;
	if (num_xtnts > QNX4FS_MAX_XTS_PER_XBLK || num_xtnts == 0) {
		QNX4FS_WARN(" wrong xblk %lu extents number %d \n", nxblk,
			num_xtnts);
		brelse(bp);
		return 0;
	}

	if (nblocks >= le32_to_cpu(xblk->xb_nblocks))
		return bp;

	blocks = nblocks;
	xtnts  = 0;

	for (xtnt = xblk->xb_xts; num_xtnts > 0; num_xtnts--) {
		if (blocks < le32_to_cpu(xtnt->size))
			break;
		blocks -= le32_to_cpu(xtnt->size);
		xtnt++;
		xtnts++;
	}

	if (blocks > 0 && num_xtnts > 0) {
		truncate_xtnt_le(sbp, xtnt, blocks);
		num_xtnts--;
		xtnts++;
		xtnt++;
	}

	while (num_xtnts-- > 0) {
		truncate_xtnt_le(sbp, xtnt, 0);
		xtnt++;
	}

	xblk->xb_nxt = xtnts;
	xblk->xb_nblocks = cpu_to_le32(nblocks);

	return bp;
}

void
qnx4fs_truncate(struct inode *ip)
{
	struct buffer_head *bp;
	struct super_block *sbp = ip->i_sb;
	struct qnx4fs_xb *xb;
	struct qnx4fs_xt *xt;
	unsigned long xbnr, xbnr_tmp;
	unsigned long isize, offset;
	unsigned long nblocks;
	int num_xtnts_rest;
	unsigned num_xtnts;

	QNX4FS_DEBUG(" inode %ld, i_size %d, i_mmu %d \n", ip->i_ino,
	(int)(ip->i_size), (int)(QNX4FS_I(ip)->i_mmu_private));

	/* this is stolen from fs/affs/file.c:affs_truncate */
	if (ip->i_size > QNX4FS_I(ip)->i_mmu_private) {
		struct address_space *mapping = ip->i_mapping;
		struct page *page;
		void *fsdata;
		u32 size = ip->i_size;

		if (!mapping->a_ops->write_begin(NULL, mapping, size, 0, 0, &page, &fsdata))
			mapping->a_ops->write_end(NULL, mapping, size, 0, 0, page, fsdata);
		mark_inode_dirty(ip);
		return;
	}

	isize = (ip->i_size + QNX4FS_BLOCK_SIZE-1) >> QNX4FS_BLOCK_SIZE_BITS;

	num_xtnts_rest = QNX4FS_I(ip)->i_nxt;
	offset = QNX4FS_I(ip)->i_xt0.size;
	xbnr = QNX4FS_I(ip)->i_xbnr0;

	if (isize <= offset) {
		qnx4fs_bfree(sbp, QNX4FS_I(ip)->i_xt0.start + isize,
			offset - isize);
		QNX4FS_I(ip)->i_xt0.size = isize;
		QNX4FS_I(ip)->i_xbnr0 = 0;
		QNX4FS_I(ip)->i_xbnr_last = 0;

		if (num_xtnts_rest && !isize) {
			QNX4FS_I(ip)->i_xt0.start = 0;
			num_xtnts_rest--;
		}
	}

	QNX4FS_I(ip)->i_offset		= 0;
	QNX4FS_I(ip)->i_offset_end	= QNX4FS_I(ip)->i_xt0.size;
	QNX4FS_I(ip)->i_phys		= QNX4FS_I(ip)->i_xt0.start;
	QNX4FS_I(ip)->i_xtnr		= -1;
	QNX4FS_I(ip)->i_xbnr		= QNX4FS_I(ip)->i_xbnr0;

	for (; xbnr; brelse(bp)) {

		bp = sb_bread(ip->i_sb, xbnr - 1);
		if (bp == NULL) {
			QNX4FS_WARN("unable to finish truncate ino %ld \n",
				ip->i_ino);
			goto out;
		}

		xb = (struct qnx4fs_xb *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			QNX4FS_WARN(" wrong xblk %lu signature \n", xbnr);
			brelse(bp);
			goto out;
		}

		num_xtnts = xb->xb_nxt;
		if (num_xtnts > QNX4FS_MAX_XTS_PER_XBLK || num_xtnts == 0) {
			QNX4FS_WARN(" wrong xblk %lu extents number %u \n",
				xbnr, num_xtnts);
			brelse(bp);
			goto out;
		}

		if (isize <= offset) {
			/* free extent-block */
			num_xtnts_rest -= num_xtnts;
			xt = xb->xb_xts;
			for (; num_xtnts > 0; num_xtnts--) {
				free_extent_le(sbp, xt);
				xt++;
			}
			xbnr_tmp = xbnr;
			xbnr = le32_to_cpu(xb->xb_next);
			qnx4fs_bfree(sbp, xbnr_tmp, 1);
			continue;
		}

		nblocks = le32_to_cpu(xb->xb_nblocks);
		offset += nblocks;
		if (isize > offset) {
			xbnr = le32_to_cpu(xb->xb_next);
			continue;
		}

		xt = &xb->xb_xts[num_xtnts - 1];
		for (; num_xtnts > 0; num_xtnts--) {
			int size = le32_to_cpu(xt->size);

			if (isize > (offset - size)) {
				/* last extent */
				size = offset - isize;
				if (size != 0)
					shorten_extent_by(sbp, xt, size);
				nblocks -= size;
				num_xtnts--;
				break;
			}
			offset  -= size;
			nblocks -= size;
			free_extent_le(sbp, xt);
			xt->start = 0;
			xt->size  = 0;
			num_xtnts_rest--;
			xt--;
		}
		offset = isize;
		xb->xb_nxt = num_xtnts + 1;
		xb->xb_nblocks = cpu_to_le32(nblocks);
		QNX4FS_I(ip)->i_xbnr_last = xbnr;
		xbnr = le32_to_cpu(xb->xb_next);
		xb->xb_next = 0;
		mark_buffer_dirty(bp);
	}
out:
	QNX4FS_I(ip)->i_nxt = num_xtnts_rest;
	QNX4FS_I(ip)->i_mmu_private = ip->i_size;
	ip->i_mtime = CURRENT_TIME;
	ip->i_ctime = CURRENT_TIME;
	mark_inode_dirty(ip);
}

